/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.component.generator.component.ComponentFunctionIdentifierUtils.getInitializeFunctionName;
import static org.fortiss.af3.component.generator.component.ComponentFunctionIdentifierUtils.getPerformStepFunctionName;
import static org.fortiss.af3.component.generator.component.PortVariableUtils.getPortNoValAssignment;
import static org.fortiss.af3.component.generator.component.PortVariableUtils.getPortPortValueAssignment;
import static org.fortiss.af3.component.generator.component.PortVariableUtils.getPortValue;
import static org.fortiss.af3.component.generator.component.PortVariableUtils.getPortValueAssignment;
import static org.fortiss.af3.component.generator.component.PortVariableUtils.testPortVariableForNoVal;
import static org.fortiss.af3.component.utils.GeneratorModelElementFactory.createComponentFunction;
import static org.fortiss.af3.component.utils.GeneratorModelElementFactory.createComponentProgram;
import static org.fortiss.af3.component.utils.GeneratorModelElementFactory.createLocalFunction;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.assignment;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.createDataDictionary;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.funcCall;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.ifthenelse;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.sequence;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFacade.addUserHeaderInclude;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFactory.createCSourcePackage;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFactory.createConfigureFile;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFactory.createMakedefsFile;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getRootElement;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.TransformationUtils.createTransformedObjectFor;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.model.generator.ComponentFunction;
import org.fortiss.af3.component.model.generator.ComponentProgram;
import org.fortiss.af3.component.model.generator.LocalFunction;
import org.fortiss.af3.expression.model.DataDictionary;
import org.fortiss.af3.expression.model.terms.IExpressionTerm;
import org.fortiss.af3.expression.model.terms.imperative.Assignment;
import org.fortiss.af3.expression.model.terms.imperative.IStatementTerm;
import org.fortiss.af3.expression.model.terms.imperative.StatementSequence;
import org.fortiss.af3.generator.common.model.c.CImplementationFile;
import org.fortiss.af3.generator.common.model.c.CSourcePackage;
import org.fortiss.af3.platform.AF3PlatformActivator;
import org.fortiss.af3.platform.language.executable.ExecutableBase;
import org.fortiss.af3.platform.language.executable.ExecutionUnitExecutableBase;
import org.fortiss.af3.platform.language.executable.IInitializableExecutable;
import org.fortiss.af3.platform.language.executable.IReadableExecutable;
import org.fortiss.af3.platform.language.executable.IReadableExecutableWithNoValSupport;
import org.fortiss.af3.platform.language.executable.ITerminatableExecutable;
import org.fortiss.af3.platform.language.executable.IWritableExecutable;
import org.fortiss.af3.platform.language.executable.IWritableExecutableWithNoValSupport;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.GenericPlatformUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.generic.GenericReceiver;
import org.fortiss.af3.platform.model.generic.GenericTransceiver;
import org.fortiss.af3.platform.model.generic.GenericTransmitter;
import org.fortiss.af3.platform.raspberry.model.Core;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;
import org.fortiss.tooling.kernel.extension.exception.TransformationFailedException;

/**
 * 
 * @author chaudhary
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public class CoreExecutable extends ExecutionUnitExecutableBase<Core, CSourcePackage> {

	/**
	 * @param modelElement
	 */
	public CoreExecutable(Core modelElement) {
		super(modelElement);
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getInitialization() {
		return funcCall("init_generic_ecu");
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getTermination() {
		return funcCall("term_generic_ecu");
	}

	/** {@inheritDoc} */
	@Override
	protected CSourcePackage createExecutionUnitSourcePackage(String name,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {

		/* need to check if all annotations are set */

		CSourcePackage sourcePackage = createCSourcePackage();
		String dirName = ((GenericPlatformUnit)modelElement.eContainer()).getName();
		sourcePackage.setBaseLocation(dirName + "/" + name);

		// Fill the the two helper sets. Sets are used to ensure uniqueness of platform connectors.
		Set<PlatformConnectorUnit> usedUnits = new HashSet<PlatformConnectorUnit>();
		Set<GenericTransceiver> usedTransceivers = new HashSet<GenericTransceiver>();
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			usedUnits.add(p.getFirst());
			if(p.getFirst() instanceof GenericTransceiver) {
				usedTransceivers.add((GenericTransceiver)p.getFirst());
			}
		}

		// Create the system functions.
		// ComponentFunction init =
		// createInitializeFunction(deployedComponents, usedUnits, deployedPorts);
		// ComponentFunction step =
		// createStepFunction(deployedComponents, deployedPorts, usedTransceivers);
		// ComponentProgram program = createComponentProgram("system", init, step);
		// program.getLocalFunctions().add(
		// createTerminationFunction(deployedPorts, deployedComponents));
		// program.getLocalFunctions().add(
		// createReadInputFunction(deployedPorts, usedTransceivers, deployedComponents));

		// Create the system functions.
		ComponentFunction init = createInitializeFunction(deployedComponents, usedUnits);
		ComponentFunction step =
				createStepFunction(deployedComponents, deployedPorts, usedTransceivers);
		ComponentProgram program = createComponentProgram("system", init, step);
		program.getLocalFunctions().add(createTerminationFunction(usedUnits));
		program.getLocalFunctions().add(createReadInputFunction(deployedPorts, usedTransceivers));

		// Fill the system program with the component sub-programs
		try {
			// deployed components
			for(Pair<ExecutionUnit, Component> pair : deployedComponents) {
				program.getSubPrograms().add(
						createTransformedObjectFor(pair.getSecond(), ComponentProgram.class,
								context));
			}

			// convert the program to C code
			CSourcePackage cPack =
					createTransformedObjectFor(program, CSourcePackage.class, context);
			cPack.mergeInto(sourcePackage);

			if(!deployedComponents.isEmpty()) {
				DataDictionary dataDict =
						getRootElement(deployedComponents.get(0).getSecond(), DataDictionary.class);
				if(dataDict == null) {
					dataDict = createDataDictionary(); // dummy element => predefined types are
					// created
				}
				CSourcePackage dataDictionaryPackage =
						createTransformedObjectFor(dataDict, CSourcePackage.class, context);
				dataDictionaryPackage.mergeInto(sourcePackage);
			}
		} catch(TransformationFailedException e) {
			error(AF3PlatformActivator.getDefault(), e.getMessage(), e);
		}

		fixSystemCImports((CImplementationFile)sourcePackage.getSrcGenPackage()
				.findSourceUnitByName("system.c"), usedUnits, usedTransceivers);

		sourcePackage.getUnits().add(createConfigureFile(modelElement.getName()));
		sourcePackage.getUnits().add(createMakedefsFile());
		return sourcePackage;
	}

	/** Adds the platform connectors includes. */
	private void fixSystemCImports(CImplementationFile system,
			Set<PlatformConnectorUnit> usedUnits, Set<GenericTransceiver> usedTransceivers) {
		// Add includes for user code of generic HAL implementation
		for(GenericTransceiver gt : usedTransceivers) {
			addUserHeaderInclude(system, gt.getName());
		}
		for(PlatformConnectorUnit unit : usedUnits) {
			if(unit instanceof GenericTransmitter || unit instanceof GenericReceiver) {
				addUserHeaderInclude(system, "io");
				break;
			}
		}
	}

	/** Creates the initialize function. */
	protected ComponentFunction createInitializeFunction(
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			Set<PlatformConnectorUnit> usedUnits) {
		List<IStatementTerm> body = new ArrayList<IStatementTerm>();
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			body.add(assignment(funcCall(getInitializeFunctionName(p.getSecond()))));
		}

		for(PlatformConnectorUnit pu : usedUnits) {
			ExecutableBase<? extends PlatformConnectorUnit> exec =
					getPlatformArchitectureExecutable().getExecutable(pu);
			if(exec instanceof IInitializableExecutable) {
				body.add(assignment(((IInitializableExecutable)exec).getInitialization()));
			}
		}
		return createComponentFunction("initialize_system", null, sequence(body));
	}

	/** Creates the termination function. */
	protected LocalFunction createTerminationFunction(Set<PlatformConnectorUnit> usedUnits) {
		List<IStatementTerm> body = new ArrayList<IStatementTerm>();
		for(PlatformConnectorUnit pu : usedUnits) {
			ExecutableBase<? extends PlatformConnectorUnit> exec =
					getPlatformArchitectureExecutable().getExecutable(pu);
			if(exec instanceof ITerminatableExecutable) {
				body.add(assignment(((ITerminatableExecutable)exec).getTermination()));
			}
		}
		return createLocalFunction("terminate_system", null, sequence(body));
	}

	/** Creates the step function. */
	protected ComponentFunction createStepFunction(
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts,
			Set<GenericTransceiver> usedTransceivers) {
		List<IStatementTerm> body = new ArrayList<IStatementTerm>();

		// Search for local channels, i.e. channels that begin and end at components deployed to
		// this generic execution unit.
		// If such channels are found, we have to place the corresponding assignments here, since
		// there is no super component that would do that for us.
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component comp = p.getSecond();
			for(InputPort conn : comp.getInputPorts()) {
				for(Channel ch : conn.getIncomingChannels()) {
					if(ch.getSource().eContainer().equals(comp)) {
						// We found a local channel; create the assignments
						body.add(getPortPortValueAssignment(ch.getSource(), ch.getTarget()));
					}
				}
			}
		}
		body.add(assignment(funcCall("read_input")));
		// Notify transceivers about start of output writing.
		for(GenericTransceiver gt : usedTransceivers) {
			body.add(assignment(funcCall("prepare_output_" + gt.getName())));
		}
		// Perform component step
		for(Pair<ExecutionUnit, Component> p : deployedComponents) {
			Component comp = p.getSecond();
			if(comp.getSubComponents().isEmpty() && comp.isStronglyCausal()) {
				writeOutputsForComponent(comp, deployedPorts, body);
			}
			body.add(assignment(funcCall(getPerformStepFunctionName(p.getSecond()))));
			if(!comp.getSubComponents().isEmpty() || !comp.isStronglyCausal()) {
				writeOutputsForComponent(comp, deployedPorts, body);
			}
		}
		// Notify transceivers about end of output writing.
		for(GenericTransceiver gt : usedTransceivers) {
			body.add(assignment(funcCall("finish_output_" + gt.getName())));
		}
		return createComponentFunction("run_system", null, sequence(body));
	}

	/**
	 * @param deployedPorts
	 * @param body
	 */
	private void writeOutputsForComponent(Component comp,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, List<IStatementTerm> body) {
		// Write outputs to transmitter and transceivers.
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			if(p.getSecond().getComponent() != comp) {
				continue;
			}
			ExecutableBase<? extends PlatformConnectorUnit> exec =
					getPlatformArchitectureExecutable().getExecutable(p.getFirst());
			if(exec instanceof IWritableExecutable && p.getSecond() instanceof OutputPort) {
				createWriteAccess((IWritableExecutable)exec, (OutputPort)p.getSecond(), body);
			}
		}
	}

	/** Creates the read input function. */
	protected LocalFunction createReadInputFunction(
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts,
			Set<GenericTransceiver> usedTransceivers) {
		List<IStatementTerm> body = new ArrayList<IStatementTerm>();
		// Notify transceivers about start input reading.
		for(GenericTransceiver gt : usedTransceivers) {
			body.add(assignment(funcCall("prepare_input_" + gt.getName())));
		}
		// Read input from transceivers and receivers.
		for(Pair<PlatformConnectorUnit, Port> p : deployedPorts) {
			ExecutableBase<? extends PlatformConnectorUnit> exec =
					getPlatformArchitectureExecutable().getExecutable(p.getFirst());
			if(exec instanceof IReadableExecutable && p.getSecond() instanceof InputPort) {
				createReadAccess((IReadableExecutable)exec, (InputPort)p.getSecond(), body);
			}
		}
		// Notify transceivers about end of input reading.
		for(GenericTransceiver gt : usedTransceivers) {
			body.add(assignment(funcCall("finish_input_" + gt.getName())));
		}
		return createLocalFunction("read_input", null, sequence(body));
	}

	/** Creates the read code for the given executable and port */
	protected void
			createReadAccess(IReadableExecutable exec, InputPort p, List<IStatementTerm> body) {
		IStatementTerm valueAssign = getPortValueAssignment(p, exec.getValueReadAccessor(p));
		if(exec instanceof IReadableExecutableWithNoValSupport) {
			IExpressionTerm guard =
					((IReadableExecutableWithNoValSupport)exec).getNoValGuardAccessor(p);
			StatementSequence thenBlock = sequence(getPortNoValAssignment(p));
			StatementSequence elseBlock = sequence(valueAssign);
			body.add(ifthenelse(guard, thenBlock, elseBlock));
		} else {
			body.add(valueAssign);
		}
	}

	/** Creates the write code for the given executable and port. */
	protected void createWriteAccess(IWritableExecutable exec, OutputPort p,
			List<IStatementTerm> body) {
		Assignment valueAssign = assignment(exec.getValueWriteAccessor(p, getPortValue(p)));
		if(exec instanceof IWritableExecutableWithNoValSupport) {
			IExpressionTerm guard = testPortVariableForNoVal(p);
			Assignment thenBlock =
					assignment(((IWritableExecutableWithNoValSupport)exec).getNoValWriteAccessor(p));
			body.add(ifthenelse(guard, sequence(thenBlock), sequence(valueAssign)));
		} else {
			body.add(valueAssign);
		}
	}

}
